from threading import Thread

from backports.asyncio import runners


class RunnerThread(Thread):

    def __init__(self, main, *, debug=False):
        super().__init__()
        self.result = None
        self.exception = None
        self.main = main
        self.debug = debug

    def run(self):
        try:
            self.result = runners.run(self.main, debug=self.debug)
        except BaseException as e:
            self.exception = e


def run(main, *, debug=False):
    """
    Since we're using asyncio loop to run wait() in irder to be compatible with async calls,
    here we also run each wait in a different thread to allow nested calls to wait()
    """
    thread = RunnerThread(main, debug=debug)
    thread.start()
    thread.join()
    if thread.exception:
        raise thread.exception
    return thread.result



import inspect
import re
from functools import partial


def is_async(func):
    return inspect.iscoroutinefunction(_unpartial(func))


def _unpartial(func):
    while isinstance(func, partial):
        func = func.func
    return func


def describe(func):
    if _is_a_lambda(func):
        return _content_of(func)
    return _unpartial(func).__name__


def _is_a_lambda(func):
    lambda_template = lambda: 0  # noqa: E731
    return isinstance(func, type(lambda_template)) and \
        func.__name__ == lambda_template.__name__


def _content_of(lambda_func):
    source_line = inspect.getsource(lambda_func)
    r = re.search(r'lambda:\s*(.+)\s*\)', source_line)
    return r.group(1)



import asyncio

import busypie
import time
from busypie.func import is_async, describe


class AsyncConditionAwaiter:
    def __init__(self, condition, func_checker):
        self._condition = condition
        self._func_check = func_checker
        self._validate_condition()

    def _validate_condition(self):
        if self._condition.poll_delay > self._condition.wait_time_in_secs:
            raise ValueError('Poll delay should be shorter than maximum wait constraint')

    async def wait_for(self, func):
        start_time = time.time()
        await asyncio.sleep(self._condition.poll_delay)
        is_func_async = is_async(func)
        while True:
            try:
                result = await self._perform_func_check(func, is_func_async)
                if result:
                    return result
            except Exception as e:
                self._raise_exception_if_not_ignored(e)
            self._validate_wait_constraint(func, start_time)
            await asyncio.sleep(self._condition.poll_interval)

    async def _perform_func_check(self, func, is_func_async):
        if is_func_async:
            return await self._func_check(func)
        else:
            return self._func_check(func)

    def _raise_exception_if_not_ignored(self, e):
        ignored_exceptions = self._condition.ignored_exceptions
        if ignored_exceptions is None or \
                (ignored_exceptions and e.__class__ not in ignored_exceptions):
            raise e

    def _validate_wait_constraint(self, condition_func, start_time):
        if (time.time() - start_time) > self._condition.wait_time_in_secs:
            raise busypie.ConditionTimeoutError(
                self._condition.description or describe(condition_func), self._condition.wait_time_in_secs)


class ConditionTimeoutError(Exception):
    def __init__(self, description, wait_time_in_secs):
        super(ConditionTimeoutError, self).__init__("Failed to meet condition of [{}] within {} seconds"
                                                    .format(description, wait_time_in_secs))
        self.description = description
